-- ***************************************************************************
-- *  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de)
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- ***************************************************************************
-- *  History:
-- *
-- *  25.12.2017  mifi  First version.
-- ***************************************************************************

-- ***************************************************************************
-- *  DEFINE: Library                                                        *
-- ***************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


-- ***************************************************************************
-- *  DEFINE: Entity                                                         *
-- ***************************************************************************

--
-- SPI <-> Wishbone Master
--
entity spi_wb_master is
   generic (
      ADDRESS_WIDTH : integer := 8
   );
   port (
      -- System  
      clk_i      : in  std_logic := '0';
      rst_i      : in  std_logic := '0';
      
      -- SPI Interface
      spi_clk_i  : in  std_logic := '0';
      spi_cs_i   : in  std_logic := '1';
      spi_mosi_i : in  std_logic := '0';
      spi_miso_o : out std_logic;
      
      -- Wishbone Master
      wbm_stb_o  : out std_logic;
      wbm_cyc_o  : out std_logic;
      wbm_we_o   : out std_logic;
      wbm_ack_i  : in  std_logic := '0';

      wbm_adr_o  : out std_logic_vector(ADDRESS_WIDTH-1 downto 0);  
      
      wbm_dat_o  : out std_logic_vector(7 downto 0);  
      wbm_dat_i  : in  std_logic_vector(7 downto 0) := (others => '0')
   );
end entity spi_wb_master;


-- ***************************************************************************
-- *  DEFINE: Architecture                                                   *
-- ***************************************************************************

architecture syn of spi_wb_master is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------
   
   constant WB_WAIT_CNT_MAX : integer := 31;
   
   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------
   
   -- SPI data state
   type   spi_state_t is ( SPI_STATE_IDLE, SPI_STATE_DATA );
   signal spi_state    : spi_state_t := SPI_STATE_IDLE;
   
   -- Command state
   type   cmd_state_t is ( CMD_STATE_IDLE, 
                           CMD_STATE_CMD, 
                           CMD_STATE_ADR, 
                           CMD_STATE_WRITEDATA,
                           CMD_STATE_MASTER_WRITE,
                           CMD_STATE_READ_COUNT,
                           CMD_STATE_MASTER_READ,
                           CMD_STATE_MASTER_READ_END
                         );
   signal cmd_state    : cmd_state_t := CMD_STATE_IDLE;
   
   -- Wishbone state
   type   wb_state_t is ( WB_STATE_IDLE, WB_STATE_READ, WB_STATE_WRITE );
   signal wb_state     : wb_state_t := WB_STATE_IDLE;
   
   -- SPI interface signals      
   signal spi_clk_sr   : std_logic_vector(2 downto 0) := (others => '0');
   signal spi_cs_sr    : std_logic_vector(2 downto 0) := (others => '1');
   signal spi_mosi_sr  : std_logic_vector(2 downto 0) := (others => '0');
   signal spi_clk      : std_logic := '0';
   signal spi_cs       : std_logic := '0';
   signal spi_mosi     : std_logic := '0';
   
   -- SPI data
   signal bit_cnt      : integer range 0 to 8         := 0;
   signal data         : std_logic_vector(7 downto 0) := (others => '0');
   signal data_valid   : std_logic                    := '0';
   
   signal mosi_sr      : std_logic_vector(7 downto 0) := (others => '0');
   signal miso_sr      : std_logic_vector(7 downto 0) := (others => '0');
   
   -- Command
   signal command      : std_logic_vector(7 downto 0) := (others => '0');
   signal address      : std_logic_vector(ADDRESS_WIDTH-1 downto 0) := (others => '0');
   signal read_cnt     : integer range 0 to 255       := 0;
   signal read_cnt_max : integer range 0 to 255       := 0;
   signal write_req    : std_logic                    := '0';
   signal read_req     : std_logic                    := '0';
   signal write_ack    : std_logic                    := '0';
   signal read_ack     : std_logic                    := '0';
   
   -- Wishbone   
   signal wb_wait_cnt  : integer range 0 to WB_WAIT_CNT_MAX := 0;  
   signal wb_stb       : std_logic                    := '0';
   signal wb_we        : std_logic                    := '0';
   signal wb_writedata : std_logic_vector(7 downto 0) := (others => '0');
	signal wb_readdata  : std_logic_vector(7 downto 0) := (others => '0');
   
   
   --
   -- Prevent the following signals to remove,
   -- needed for SingnalTap debugging.
   --
   attribute noprune              : boolean;
   attribute noprune of spi_cs    : signal is true;
   attribute noprune of spi_mosi  : signal is true;
   
begin

   --
   -- Synchronize input signals
   --  
   process (clk_i)
   begin
      if rising_edge(clk_i) then
         if (rst_i = '1') then
            -- Synchronous RESET
            spi_cs_sr   <= (others => '1');
            spi_clk_sr  <= (others => '0');
         else
            spi_cs_sr   <= spi_cs_sr(1 downto 0)   & NOT spi_cs_i;
            spi_clk_sr  <= spi_clk_sr(1 downto 0)  & spi_clk_i;
            spi_mosi_sr <= spi_mosi_sr(1 downto 0) & spi_mosi_i;
         end if;
      end if;
   end process;

   spi_cs   <= spi_cs_sr(2);
   spi_mosi <= spi_mosi_sr(2);


   --
   -- Single Process FSM, data processing
   -- 
   process (clk_i)
   begin
      if rising_edge(clk_i) then
         if (rst_i = '1') then   
            -- Synchronous RESET 
            spi_state <= SPI_STATE_IDLE;
         else
         
            -- Default
            spi_clk    <= '0';
            data_valid <= '0';
            
            
            -- SPI Clock, check for rising edge 
            if ((spi_clk_sr(2) = '0') and (spi_clk_sr(1) = '1')) then
               spi_clk <= '1';
            end if;

            
            -- FSM start
            case (spi_state) is
            
               when SPI_STATE_IDLE =>
                  bit_cnt <= 0;
               
                  if (spi_cs = '1') then
                     spi_state <= SPI_STATE_DATA;
                  end if;
               -- SPI_STATE_IDLE
               
               
               when SPI_STATE_DATA =>
                  if (spi_cs = '1') then
                  
                     if (spi_clk = '1') then
                        mosi_sr <= mosi_sr(6 downto 0) & spi_mosi;   -- Shift data in      
                        miso_sr <= miso_sr(6 downto 0) & '0';        -- Shift data out
                        
                        if (bit_cnt < 7) then
                           bit_cnt <= bit_cnt + 1;
                        else
                           bit_cnt    <= 0;
                           
                           data       <= mosi_sr(6 downto 0) & spi_mosi;
                           data_valid <= '1';                        -- Signal data in valid
                           miso_sr    <= wb_readdata;                -- Next data to shift out
                        end if;   
                        
                     end if;
                  
                  else
                     spi_state <= SPI_STATE_IDLE;   
                  end if;
               -- SPI_STATE_DATA
               
               
               when others =>
                  spi_state <= SPI_STATE_IDLE;
               -- Others   

            end case; 
            
         end if; -- End "rst_i"
      end if; -- End "clk_i"
   end process;


   --
   -- Single Process FSM, command processing
   -- 
   process (clk_i)
   begin
      if rising_edge(clk_i) then
         -- Synchronous RESET 
         if (rst_i = '1') then   
            cmd_state <= CMD_STATE_IDLE;
         else
            
            -- Default
            write_req <= '0';
            read_req  <= '0';

            -- FSM start
            case (cmd_state) is
            
               when CMD_STATE_IDLE =>
                  if (spi_cs = '1') then
                     cmd_state <= CMD_STATE_CMD;
                  end if;
               -- CMD_STATE_IDLE

               
               when CMD_STATE_CMD =>
                  if (spi_cs = '1') then
                     if (data_valid = '1') then                   -- Wait for data valid
                        command   <= data;                        -- Command and addres-hi data
                        cmd_state <= CMD_STATE_ADR;               -- Switch to address state
                     end if;
                  else
                     cmd_state <= CMD_STATE_IDLE;                 -- Ups, chip select was removed
                  end if;
               -- CMD_STATE_CMD
               
               
               when CMD_STATE_ADR =>
                  if (spi_cs = '1') then
                     if (data_valid = '1') then                   -- Wait for data valid
                        address <= command(4 downto 0) & data;    -- Address-lo data
                        
                        if (command(7 downto 5) = "001") then     -- Check for write command
                           cmd_state <= CMD_STATE_WRITEDATA;
                        end if;   
                        
                        if (command(7 downto 5) = "010") then     -- Check for read command
                           cmd_state <= CMD_STATE_READ_COUNT;
                        end if;   
                        
                     end if;
                  else
                     cmd_state <= CMD_STATE_IDLE;
                  end if;
               -- CMD_STATE_ADR
               
               
               when CMD_STATE_WRITEDATA =>
                  if (spi_cs = '1') then
                     if (data_valid = '1') then                   -- Wait for data valid
                        wb_writedata <= data;                     -- Data to write by the Wishbone bus
                        write_req    <= '1';                      -- Signal write request
                        cmd_state    <= CMD_STATE_MASTER_WRITE;
                     end if;
                  else
                     cmd_state <= CMD_STATE_IDLE;
                  end if;
               -- CMD_STATE_WRITEDATA
               
               
               when CMD_STATE_MASTER_WRITE =>
                  if (spi_cs = '1') then
                     if (write_ack = '1') then                                -- Wait for write ready
                        address   <= std_logic_vector(unsigned(address) + 1); -- Autoincrement address for the next round 
                        cmd_state <= CMD_STATE_WRITEDATA;
                     end if;
                  else
                     cmd_state <= CMD_STATE_IDLE;
                  end if;
               -- CMD_STATE_MASTER_WRITE
               
               
               when CMD_STATE_READ_COUNT =>
                  if (spi_cs = '1') then
                     if ((spi_clk = '1') and (bit_cnt = 0)) then     -- Start read request by the first bit
                        read_req <= '1';                             -- Signal read request
                     end if;
                     
                     if (read_ack = '1') then                                 -- Wait for read ready
                        address <= std_logic_vector(unsigned(address) + 1);   -- Autoincrement address for the next round
                     end if;   
                     
                     if (data_valid = '1') then                      -- Wait for data valid from the SPI bus
                        read_cnt     <= 0;                           -- Set read count to 0
                        read_cnt_max <= to_integer(unsigned(data));  -- Get requested read count = (Count-1)
                        
                        if (data = x"00") then                       -- 0 means only one byte to read
                           cmd_state <= CMD_STATE_MASTER_READ_END;
                        else
                           cmd_state <= CMD_STATE_MASTER_READ;       -- Read more than one byte
                        end if;   
                     end if;   
                  else
                     cmd_state <= CMD_STATE_IDLE;
                  end if;
               -- CMD_STATE_READ_COUNT
               
               
               when CMD_STATE_MASTER_READ =>
                  if (spi_cs = '1') then
                  
                     if ((spi_clk = '1') and (bit_cnt = 0)) then     -- Start read request by the first bit
                        if (read_cnt < read_cnt_max) then
                           read_cnt <= read_cnt + 1;
                           read_req <= '1';                          -- Signal read request
                        else   
                           cmd_state <= CMD_STATE_MASTER_READ_END;   -- No more bytes to read
                        end if;
                     end if;
                     
                     if (read_ack = '1') then                                 -- Wait for read ready
                        address <= std_logic_vector(unsigned(address) + 1);   -- Autoincrement address for the next round
                     end if;   
                  else
                     cmd_state <= CMD_STATE_IDLE;
                  end if;
               -- CMD_STATE_MASTER_READ
               

               when CMD_STATE_MASTER_READ_END =>
                  if (spi_cs = '1') then
                     -- Do nothing
                  else
                     cmd_state <= CMD_STATE_IDLE;
                  end if;
               -- CMD_STATE_MASTER_READ_END
               
               
               when others =>
                  cmd_state <= CMD_STATE_IDLE;
               -- Others   

            end case;

         end if;
         
      end if;
   end process;
   
   
   --
   -- Single Process FSM
   -- Used to support the following master signals: STB_O, WE_O and ACK_I
   -- 
   process (clk_i, rst_i)     
   begin
      if (rst_i = '1') then
         wb_state <= WB_STATE_IDLE;
         wb_stb   <= '0';
         wb_we    <= '0';
        
      elsif rising_edge(clk_i) then
                     
         -- Default
         write_ack <= '0';
         read_ack  <= '0';
         wb_stb    <= wb_stb;
         wb_we     <= wb_we;
         
         case (wb_state) is
         
            when WB_STATE_IDLE =>
               wb_stb      <= '0';
               wb_we       <= '0';
               wb_wait_cnt <= 0;
               
               if (write_req = '1') then        -- Check for write request
                  wb_stb   <= '1';              -- Signal valid data transfer cycle on the Wishbone bus
                  wb_we    <= '1';              -- Signal a write cycle on the Wishbone bus
                  wb_state <= WB_STATE_WRITE;
               end if;

               if (read_req = '1') then         -- Check for read request
                  wb_stb   <= '1';              -- Signal valid data transfer cycle on the Wishbone bus
                  wb_we    <= '0';              -- Signal a read cycle on the Wishbone bus
                  wb_state <= WB_STATE_READ;
               end if;
            -- WB_STATE_IDLE
            
            
            when WB_STATE_WRITE =>                    
               if (wbm_ack_i = '1') then        -- Wait for a normal termination of a Wishbone bus cycle
                  wb_stb    <= '0';             -- Signal end of data transfer on the Wishbone bus
                  wb_we     <= '0';             -- Not needed, but switch back to read mode
                  write_ack <= '1';
                  wb_state  <= WB_STATE_IDLE;
               else
                  if (wb_wait_cnt < WB_WAIT_CNT_MAX) then
                     wb_wait_cnt <= wb_wait_cnt + 1;             
                  end if;   
                  wb_state <= WB_STATE_WRITE;
               end if;
            -- WB_STATE_WRITE
            
            
            when WB_STATE_READ =>               
               if (wbm_ack_i = '1') then        -- Wait for a normal termination of a Wishbone bus cycle                  
                  wb_stb      <= '0';           -- Signal end of data transfer on the Wishbone bus
                  wb_readdata <= wbm_dat_i;     -- Wishbone data can be stored
                  read_ack    <= '1';
                  wb_state    <= WB_STATE_IDLE;
               else  
                  if (wb_wait_cnt < WB_WAIT_CNT_MAX) then
                     wb_wait_cnt <= wb_wait_cnt + 1;             
                  end if;   
                  wb_state <= WB_STATE_READ;
               end if;
            -- WB_STATE_READ
            
            
            when others =>
               wb_state <= WB_STATE_IDLE;
            -- Others    
               
         end case;
         
         --
         -- Abort the cycle if the bus is occupied too long.
         --
         if (wb_wait_cnt = WB_WAIT_CNT_MAX) then
            wb_state <= WB_STATE_IDLE;
         end if;
         
      end if;
   end process;

      
   --------------------------------------------------------
   -- Output
   --------------------------------------------------------
   
   -- PERMISSION 3.40
   -- If a MASTER doesn't generate wait states, 
   -- then [STB_O] and [CYC_O] MAY be assigned the same signal.
   wbm_stb_o  <= wb_stb;
   wbm_cyc_o  <= wb_stb;
   
   wbm_we_o   <= wb_we;

   -- Address
   wbm_adr_o  <= address;
   
   -- Data
   wbm_dat_o  <= wb_writedata;  
   spi_miso_o <= miso_sr(7);

end architecture syn;

-- *** EOF ***
