-- ***************************************************************************
-- *  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de)
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- ***************************************************************************
-- *  History:
-- *
-- *  27.12.2017  mifi  First version.
-- ***************************************************************************

-- ***************************************************************************
-- *  DEFINE: Library                                                        *
-- ***************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


-- ***************************************************************************
-- *  DEFINE: Entity                                                         *
-- ***************************************************************************

entity wb_timer is
   generic (
      DATA_WIDTH : integer := 8
   );
   port (  
      -- Syscon
      clk_i    : in  std_logic := '0';
      rst_i    : in  std_logic := '0';
      
      -- Wishbone
      wb_stb_i : in  std_logic := '0';
      wb_cyc_i : in  std_logic := '0';      -- Not used here
      wb_we_i  : in  std_logic := '0';
      wb_ack_o : out std_logic;
      wb_adr_i : in  std_logic_vector(2 downto 0)            := (others => '0');
      wb_dat_i : in  std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
      wb_dat_o : out std_logic_vector(DATA_WIDTH-1 downto 0);  
            
      -- None Avalon
      zero_o   : out std_logic;
      int_o    : out std_logic
   );
end entity wb_timer;


-- *****************************************************************************
-- *  DEFINE: Architecture                                                     *
-- *****************************************************************************

architecture syn of wb_timer is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------
   
   constant ADR_STATUS   : std_logic_vector(2 downto 0) := "000"; -- Read / Write
   constant ADR_CONTROL  : std_logic_vector(2 downto 0) := "001"; -- Read / Write
   constant ADR_PERIOD_0 : std_logic_vector(2 downto 0) := "010"; -- Read / Write
   constant ADR_PERIOD_1 : std_logic_vector(2 downto 0) := "011"; -- Read / Write
   constant ADR_PERIOD_2 : std_logic_vector(2 downto 0) := "100"; -- Read / Write
   constant ADR_PERIOD_3 : std_logic_vector(2 downto 0) := "101"; -- Read / Write
   
   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------
   
   signal status     : std_logic_vector(7 downto 0)  := (others => '0');  
   signal control    : std_logic_vector(7 downto 0)  := (others => '0');  
   signal reload     : std_logic_vector(31 downto 0) := (others => '0');
   signal counter    : std_logic_vector(31 downto 0) := (others => '0');
   signal timer_en   : std_logic := '0';
   signal timer_zero : std_logic := '0';
   signal timer_int  : std_logic := '0';
   
   signal readdata   : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   signal read_mux   : std_logic_vector(DATA_WIDTH-1 downto 0) := (others => '0');
   
   signal read_req   : std_logic := '0'; 
   signal read_ack   : std_logic := '0';
   
   signal write_req  : std_logic := '0';
   signal write_ack  : std_logic := '0';
   
begin

   -- Some simplifications for later use
   read_req  <= '1' when ((wb_stb_i = '1') and (wb_we_i = '0')) else '0';
   write_req <= '1' when ((wb_stb_i = '1') and (wb_we_i = '1')) else '0';

   --
   -- Register File
   --
   process (clk_i, rst_i)
   begin
      if (rst_i = '1') then

         control   <= (others => '0');
         write_ack <= '0';
      
      elsif rising_edge(clk_i) then
      
         -- Default
         write_ack <= '0';
      
         if (write_req = '1') then
         
            write_ack <= '1';    -- Signal normal Wishbone bus cycle termination
         
            -- Address decoder
            case wb_adr_i is
            
               when ADR_CONTROL  => control <= wb_dat_i;
               when ADR_PERIOD_0 => reload( 7 downto  0) <= wb_dat_i;
               when ADR_PERIOD_1 => reload(15 downto  8) <= wb_dat_i;
               when ADR_PERIOD_2 => reload(23 downto 16) <= wb_dat_i;
               when ADR_PERIOD_3 => reload(31 downto 24) <= wb_dat_i;

               when others =>
                  -- Do nothing
            
            end case;
         end if;
         
      end if; 
   end process;
   
   
   --
   -- Timer
   --
   timer_en <= control(0);

   process (clk_i)
   begin
      if rising_edge(clk_i) then
      
         if (timer_en = '1') then
            if (counter = x"00000000") then
               counter    <= reload;
               timer_zero <= NOT timer_zero;
               timer_int  <= '1';
            else
               counter <= std_logic_vector(unsigned(counter) - 1);
            end if;
         else
            timer_zero <= '0';   
            timer_int  <= '0';
         end if;
         
         -- Clear timer_int on STATUS write
         if ((write_req = '1') and (wb_adr_i = ADR_STATUS)) then
            timer_int <= '0';
         end if;
         
      end if;
   end process;      
   
   
   --
   -- Read Multiplexer
   --
   status(0) <= timer_int;
   
   read_mux <= status               when (ADR_STATUS   = wb_adr_i) else 
               control              when (ADR_CONTROL  = wb_adr_i) else
               reload( 7 downto  0) when (ADR_PERIOD_0 = wb_adr_i) else
               reload(15 downto  8) when (ADR_PERIOD_1 = wb_adr_i) else
               reload(23 downto 16) when (ADR_PERIOD_2 = wb_adr_i) else
               reload(31 downto 24) when (ADR_PERIOD_3 = wb_adr_i) else (others => '0');
               

   process (clk_i)
   begin
      if rising_edge(clk_i) then
         -- Default
         read_ack <= '0';        
      
         if (read_req = '1') then
            readdata <= read_mux;
            read_ack <= '1';     -- Signal normal Wishbone bus cycle termination
         end if;
      end if;
   end process;      

   
   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

   -- Wishbone
   wb_ack_o <= read_ack or write_ack;
   wb_dat_o <= readdata;

   -- None Wishbone
   zero_o <= timer_zero;
   int_o  <= timer_int;
   
end architecture syn;

-- *** EOF ***

